// -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*-
//
// kw_catcher, Visit: www.pdfhacks.com/kw_index/
//

#include <iostream>
#include <fstream>
#include <string>
#include <map>
#include <vector>
#include <set>

using namespace std;

// this isn't an exclusion list; rather, we use stop words to detect
// the 'verbosity' of a page for weighting purposes;
//
static bool
is_stop_word( string& ss ) {
  return ( ss.length()<= 4 &&
           ( ss== "the" ||
             ss== "of" ||
             ss== "and" ||
             ss== "to" ||
             ss== "in" ||
             ss== "that" ||
             ss== "is" ||
             ss== "was" ||
             ss== "for" ||
             ss== "with" ) );
}

// pack whitespace-delimited words from an input stream into
// a couple convenient containers
//
static void
collect_words( istream& data,
               vector< map< string, unsigned long > >& page_word_count,
               map< string, unsigned long >& doc_word_count )
{
  string word;
  unsigned long page_num= 0;

  page_word_count.push_back( map< string, unsigned long >() );

  while( data ) {
    char cc= data.get();

    if( isspace(cc) || !data ) { // done with word

      // create a copy of word, removing any non-alpha chars from word ends
      string word_copy;
      for( string::iterator sit= word.begin(); sit!= word.end(); ++sit ) {
        if( isalpha( *sit ) ) {
          word_copy.assign( sit, word.end() );
          break;
        }
      }
      string::size_type word_copy_len= word_copy.length();
      while( word_copy_len&& !isalpha( word_copy[word_copy_len- 1] ) ) {
        word_copy.erase(word_copy_len- 1);
        --word_copy_len;
      }

      if( 1< word_copy.length() ) {
        if( is_stop_word( word_copy ) ) {
          // note special handling of our stop words
          (page_word_count.back())["STOPWORD"]++;
          doc_word_count["STOPWORD"]++;
        }
        else {
          (page_word_count.back())[word_copy]++;
          doc_word_count[word_copy]++;
        }
      }
      word.erase();

      // done with page?
      if( cc== 0x0c ) { // formfeed; remember: FF is whitespace
        page_word_count.push_back( map< string, unsigned long >() );
        ++page_num;
      }
    }
    else { // downcase and append to word
      word.append( 1, (char)tolower(cc) );
    }
  }

  // drop trailing empty
  if( page_word_count.back().empty() ) {
    page_word_count.pop_back();
  }
}

// sum weighted pages from page_word_count into page_window_count
//
static void
fill_window( const vector< map< string, unsigned long > >& page_word_count,
             map< string, unsigned long >& page_window_count,
             unsigned long window_start,
             int window_size )
{
  if( page_word_count.size()< window_start+ window_size ) { // error
    cerr << "error: fill_window overrun" << endl;
    return;
  }

  // iterate over pages in the window
  for( unsigned long ii= window_start; ii< window_start+ window_size; ++ii ) {

    // what proportion of the document does this page represent? use our
    // stopword count as an measure of page 'verbosity'
    //
    unsigned long stopword_weight= 0; {
      map< string, unsigned long >::const_iterator jt= page_word_count[ii].find("STOPWORD");
      if( jt!= page_word_count[ii].end() ) {
        stopword_weight= jt->second;
      }
    }
    
    // add the weighted word counts from this page to the larger window
    for( map< string, unsigned long >::const_iterator it= page_word_count[ii].begin();
         it!= page_word_count[ii].end(); ++it )
      {
        page_window_count[ it->first ]+= stopword_weight* it->second;
      }
  }
}

// identify document keywords
//
static void
select_words( const vector< map< string, unsigned long > >& page_word_count,
              const map< string, unsigned long >& doc_word_count,
              map< unsigned long, set< string > >& count_keywords,
              set< string >& keywords,
              int window_size )
{
  unsigned long num_pages= page_word_count.size();

  // by 'windowing' the entire document, we can compare local window
  // word frequency to document 'window' word frequency and be sure
  // of comparing apples to apples; note that page data gets weighted
  // when computing the window;
  // 
  map< string, unsigned long > doc_window_word_count; {
    fill_window( page_word_count,
                 doc_window_word_count,
                 0, num_pages );
  }

  // iterate over all possible document windows
  for( int ii= 0; ii< num_pages- window_size+ 1; ++ii ) {
    map< string, unsigned long > page_window_count;
    fill_window( page_word_count, page_window_count, ii, window_size );

    // iterate over all words in the window
    for( map< string, unsigned long >::const_iterator it= page_window_count.begin();
         it!= page_window_count.end(); ++it )
      {
        map< string, unsigned long >::const_iterator jt= doc_word_count.find(it->first);
        if( jt!= doc_word_count.end() && 

            // ignore keywords with low counts
            2< jt->second ) 
          {
            // if most of the word occurrences are in this window, then
            // the word appears to be a document keyword; make this
            // statistically sound by using confidence testing [TODO]
            //
            if( 0.65 < 1.0* it->second/ doc_window_word_count[it->first] ) {
              count_keywords[jt->second].insert( it->first );
              keywords.insert( it->first );
            }
          }
      }
  }
}

// output keywords in the same order they appeared in input stream
//
static void
output_in_order( istream& data,
                 ostream& output,
                 set< string > keywords,
                 bool repeat_keywords_b )
{
  string word;
  string word_prev;
  bool first_word_b= true;

  while( data ) {
    char cc= data.get();

    if( isspace(cc) || !data ) { // done with word

      // remove any non-alpha chars from word ends
      string word_copy;
      for( string::iterator sit= word.begin(); sit!= word.end(); ++sit ) {
        if( isalpha( *sit ) ) {
          word_copy.assign( sit, word.end() );
          break;
        }
      }
      string::size_type word_copy_len= word_copy.length();
      while( word_copy_len&& !isalpha( word_copy[word_copy_len- 1] ) ) {
        word_copy.erase(word_copy_len- 1);
        --word_copy_len;
      }

      // don't ever duplicate adjacent words
      if( word_copy!= word_prev ) {
        set< string >::iterator it= keywords.find( word_copy );
        if( it!= keywords.end() ) {
          if( !first_word_b ) {
            output << " ";
          }
          output << word_copy;
          if( !repeat_keywords_b ) {
            keywords.erase( it );
          }
          first_word_b= false;
          word_prev= word_copy;
        }
      }

      word.erase();
    }
    else { // append to word
      word.append( 1, (char)tolower(cc) );
    }
  }
}

int main( int argc, char** argv ) {

  vector< map< string, unsigned long > > page_word_count;
  map< string, unsigned long > doc_word_count;
  map< unsigned long, set< string > > count_keywords;
  set< string > keywords;

  bool success_b= false;

  if( argc== 4 ) {
    string report_type= argv[2];

    if( report_type== "keywords_only" ||
        report_type== "frequency" ||
        report_type== "reading_order" ||
        report_type== "reading_order_repeat" ||
        report_type== "all_reports" ) 
      {
        ifstream input_file_stream;
        bool input_file_stream_b= false;
        if( strcmp( argv[3], "-" )!= 0 ) {
          input_file_stream.open( argv[3], ios_base::in );
          input_file_stream_b= true;
        }
        istream& input_data= (input_file_stream_b) ? input_file_stream : cin;

        if( input_data ) {
          collect_words( input_data, 
                         page_word_count, 
                         doc_word_count );

          int window_size= atoi( argv[1] );
          if( window_size< 1 ||
              doc_word_count.size()< window_size )
            {
              window_size= 2;
              cerr << "warning: invalid window size; defaulting to " << window_size << endl;
            }

          select_words( page_word_count,
                        doc_word_count,
                        count_keywords,
                        keywords,
                        window_size );

          // report input settings
          cout << "#arguments - input file: " << argv[3] << " - window size: " << window_size << endl;

          if( report_type== "keywords_only" ) { // not output with all_reports
            cout << "#keywords_only" << endl;
            for( map<unsigned long, set<string> >::reverse_iterator it= count_keywords.rbegin();
                 it!= count_keywords.rend(); ++it )
              {
                for( set<string>::iterator jt= it->second.begin();
                     jt!= it->second.end(); ++jt )
                  {
                    cout << *jt << endl;
                  }
              }
            cout << endl;
          }

          // show the word frequency of our selected keywords
          if( report_type== "frequency" || report_type== "all_reports" ) {
            cout << "#frequency (keyword frequency)" << endl;
            for( map<unsigned long, set<string> >::reverse_iterator it= count_keywords.rbegin();
                 it!= count_keywords.rend(); ++it )
              {
                cout << it->first << " ";
                for( set<string>::iterator jt= it->second.begin();
                     jt!= it->second.end(); ++jt )
                  {
                    cout << *jt << " ";
                  }
                cout << endl;
              }
            cout << endl;
          }

          ////
          // output keywords by reading order to preserve
          // possible phrases and contexts
          //
          if( report_type== "reading_order" || report_type== "all_reports" ) {
            // do not repeat keywords
            cout << "#reading_order (keywords listed in reading order, with no repeats)" << endl;
            input_data.clear();
            input_data.seekg( 0, ios_base::beg );
            input_data.sync();

            output_in_order( input_data, cout, keywords, false );

            cout << endl << endl;
          }
          //
          if( report_type== "reading_order_repeat" || report_type== "all_reports" ) {
            // repeated keywords
            cout << "#reading_order_repeat (keywords listed in reading order ~with~ repeats)" << endl;
            input_data.clear();
            input_data.seekg( 0, ios_base::beg );
            input_data.sync();

            output_in_order( input_data, cout, keywords, true );

            cout << endl << endl;
          }

          success_b= true;
        }
        else { // error
          cerr << "error: unable to open file: " << argv[3] << endl;
        }
      }
  }

  if( !success_b ) {
    cerr << "kw_catcher 1.0, Identifies Keywords from a Doc's Full Text" << endl;
    cerr << endl;
    cerr << "   kw_catcher requires three arguments: " << endl;
    cerr << "      <window size> <report style> <filename>" << endl;
    cerr << "   where: " << endl;
    cerr << "      <window size> is given in pages," << endl;
    cerr << "      <report style> is:" << endl;
    cerr << "         keywords_only," << endl;
    cerr << "         frequency," << endl;
    cerr << "         reading_order," << endl;
    cerr << "         reading_order_repeat," << endl;
    cerr << "         all_reports," << endl;
    cerr << "      <filename> is the name of a text file with formfeed (0x0c)" << endl;
    cerr << "         page delimiters, as created by xpdf's pdftotext, for example." << endl;
    cerr << "         When processing data from stdin, use a hyphen (-) for the filename." << endl;
    cerr << endl;
    cerr << "   See Also: xpdf's pdftotext, page_refs";
    cerr << endl;

    return 1;
  }
  else {
    return 0;
  }
}

/* used for initial investigations

// find and return the maximum number of stop words on any page
//
static unsigned long
max_stop_word_count( const vector< map< string, unsigned long > >& page_word_count )
{
  unsigned long ret_val= 0;
  for( vector< map< string, unsigned long > >::const_iterator it= page_word_count.begin();
       it!= page_word_count.end(); ++it )
    {
      map< string, unsigned long >::const_iterator jt= it->find("STOPWORD");
      if( jt!= it->end() &&
          ret_val< jt->second )
        {
          ret_val= jt->second;
        }
    }
  return ret_val;
}

static void
report( ostream& output, 
        const vector< map< string, unsigned long > >& page_word_count )
{
  unsigned long page_num= 0;
  unsigned long max_stop_count= max_stop_word_count( page_word_count );
  for( vector< map< string, unsigned long > >::const_iterator it= page_word_count.begin();
       it!= page_word_count.end(); ++it )
    {
      output << ++page_num;

      map< string, unsigned long >::const_iterator jt= it->find("STOPWORD");
      if( jt!= it->end() ) {
        int num_hashes= 40* jt->second/ max_stop_count;
        for( int ii= 0; ii< num_hashes; ++ii ) {
          output << "#";
        }
        output << endl;
      }
      else {
        output << "0" << endl;
      }
    }  
}
*/
