// -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*-
//
// page_refs, Visit: www.pdfhacks.com/kw_index/
//

#include <iostream>
#include <fstream>
#include <string>
#include <sstream>
#include <vector>
#include <map>
#include <set>

using namespace std;

// assemble keywords into phrases if possible and assign page numbers
//
static void
compile_index( istream& data,
               map< string, set< unsigned long > >& keyword_index,
               const set< string >& keywords )
{
  string word;
  string word_prev;
  string phrase;
  unsigned long page_num= 0;
  unsigned long phrase_page_num= 0; // in case a phrase spans two pages

  while( data ) {
    char cc= data.get();

    if( isspace(cc) || !data ) { // done with word

      // create a copy of word, removing any non-alpha chars from word ends
      string word_copy;
      for( string::iterator sit= word.begin(); sit!= word.end(); ++sit ) {
        if( isalpha( *sit ) ) {
          word_copy.assign( sit, word.end() );
          break;
        }
      }
      string::size_type word_copy_len= word_copy.length();
      while( word_copy_len&& !isalpha( word_copy[word_copy_len- 1] ) ) {
        word_copy.erase(word_copy_len- 1);
        --word_copy_len;
      }

      // is word_copy a keyword?
      set< string >::const_iterator it= keywords.find( word_copy );
      if( it== keywords.end() ) { // not a keyword
        if( !phrase.empty() ) { // tally our collected phrase
          keyword_index[phrase].insert( phrase_page_num );

          phrase.erase();
        }
      }
      else if( word_copy!= word_prev ) { // word_copy is a (new) keyword; append to phrase
        if( phrase.empty() ) { // record page where phrase begins
          phrase_page_num= page_num;
        }
        else {
          phrase+= " ";
        }

        phrase+= word_copy;
        word_prev= word_copy;
      }

      word.erase();

      // done with page?
      if( cc== 0x0c ) { // formfeed; remember: FF is whitespace
        ++page_num;
      }
    }
    else{ // append cc to word
      word.append( 1, (char)tolower(cc) );
    }
  }

  // catch final phrase
  if( !phrase.empty() ) { // tally our collected phrase
    keyword_index[phrase].insert( phrase_page_num );
  }

}

string
latin_letters( int ii,
               bool downcase_b )
{
  ii-= 1; // convert from 1- to zero-based

  char cc= ii % 26;
  cc+= ((downcase_b) ? 'a' : 'A');

  int xx= ii / 26 + 1;
  return string( xx, cc );
}

string
roman_numeral( int ii,
               bool downcase_b )
  // convert input integer to a roman numeral;
  // input must be less than 5000;
  // Refs:
  // http://home.hiwaay.net/~lkseitz/math/roman/numerals.shtml
  // http://www.unc.edu/~rowlett/units/roman.html
{
  string ss;
  int jj= 0;
  ii= ii % 5000; // no numbers greater than 4999

  jj= ii / 1000;
  ii= ii % 1000;
  ss.append( jj, 'M' );
  
  jj= ii / 100;
  ii= ii % 100;
  switch( jj ) {
  case 9: // 900
    ss.append( "CM" );
    break;
  case 4: // 400
    ss.append( "CD" );
    break;
  default:
    if( 5<= jj ) {
      ss.append( "D" );
      jj-= 5;
    }
    ss.append( jj, 'C' );
  }

  jj= ii / 10;
  ii= ii % 10;
  switch( jj ) {
  case 9: // 90
    ss.append( "XC" );
    break;
  case 4: // 40
    ss.append( "XL" );
    break;
  default:
    if( 5<= jj ) {
      ss.append( "L" );
      jj-= 5;
    }
    ss.append( jj, 'X' );
  }

  jj= ii;
  switch( jj ) {
  case 9: // 9
    ss.append( "IX" );
    break;
  case 4: // 4
    ss.append( "IV" );
    break;
  default:
    if( 5<= jj ) {
      ss.append( "V" );
      jj-= 5;
    }
    ss.append( jj, 'I' );
  }

  if( downcase_b ) {
    for( string::iterator it= ss.begin(); it!= ss.end(); ++it ) {
      *it= tolower(*it);
    }
  }

  return ss;
}

int main( int argc, char** argv ) {

  bool success_b= false;

  if( argc== 3 || argc== 4 ) {

    // load the keywords from file or stdin
    set< string > keywords; {
      ifstream keyword_file_stream;
      bool keyword_file_stream_b= false;
      if( strcmp( argv[2], "-" )!= 0 ) {
        keyword_file_stream.open( argv[2], ios_base::in );
        keyword_file_stream_b= true;
      }
      istream& keyword_data= (keyword_file_stream_b) ? keyword_file_stream : cin;

      string ss;
      while( keyword_data ) {
        getline( keyword_data, ss );
        
        if( !ss.empty() && ss[0]!= '#' ) {
          // downcase
          for( string::iterator it= ss.begin(); it!= ss.end(); ++it ) {
            *it= tolower(*it);
          }
          keywords.insert( ss );
        }
      }
    }

    // load the document data, if given
    bool page_labels_b= false;
    vector< string > page_labels; {
      if( argc== 4 ) {
        ifstream doc_data_fp( argv[3], ios_base::in );
        if( doc_data_fp ) {
          long num_pages= 0;

          string current_style= "DecimalArabicNumerals";
          string current_prefix= "";
          long current_page_index= 1;
          long current_page_label= 1;

          string key, value;
          while( doc_data_fp ) {
            doc_data_fp >> key;
            if( key== "NumberOfPages:" ) {
              doc_data_fp >> value;
              num_pages= atoi( value.c_str() );
            }
            else if( key== "PageLabelNewIndex:" || !doc_data_fp ) {
              // the 1-based index where the new changes begin
              // use current state to label pages up to this index

              long new_label_index= 0;
              if( doc_data_fp ) {
                doc_data_fp >> value;
                new_label_index= atoi( value.c_str() );
              }
              else { // no more label data; number rest of document
                new_label_index= num_pages+ 1;
              }

              for( int ii= current_page_index; ii< new_label_index; ++ii ) {
                ostringstream label;
                label << current_prefix;
                if( current_style== "DecimalArabicNumerals" ) {
                  label << current_page_label++;
                }
                else if( current_style== "UppercaseRomanNumerals" ) {
                  label << roman_numeral( current_page_label++, false );
                }
                else if( current_style== "LowercaseRomanNumerals" ) {
                  label << roman_numeral( current_page_label++, true );
                }
                else if( current_style== "UppercaseLetters" ) {
                  label << latin_letters( current_page_label++, false );
                }
                else if( current_style== "LowercaseLetters" ) { // TODO
                  label << latin_letters( current_page_label++, true );
                }
                else if( current_style== "NoNumber" ) {
                  current_page_label++;
                }
                else { // error
                  cerr << "Error: page label style not given; defaulting to numbers" << endl;
                  label << current_page_label++;
                }

                page_labels.push_back( label.str() );
                // cout << label.str() << endl; // debug
              }
              current_page_index= new_label_index;

              // reset prefix, which is empty by default
              current_prefix.erase();
            }
            else if( key== "PageLabelStart:" ) {
              doc_data_fp >> value;
              current_page_label= atoi( value.c_str() );
            }
            else if( key== "PageLabelPrefix:" ) {
              doc_data_fp >> current_prefix;
            }
            else if( key== "PageLabelNumStyle:" ) {
              doc_data_fp >> current_style;
            }
            else {
              getline( doc_data_fp, value ); // skip
            }
          }

          // were we successful?
          page_labels_b= ( num_pages && page_labels.size()== num_pages );
          if( !page_labels_b ) {
            cerr << "Warning: page labels not parsed from data file." << endl;
            cerr << "Num Pages: " << num_pages << ", Num Labels: " << page_labels.size() << endl;
          }
        }
      }
    }

    // the document's scraped text
    ifstream input_file_stream;
    input_file_stream.open( argv[1], ios_base::in );
    istream& input_data= input_file_stream;
    
    if( input_data ) {

      map< string, set< unsigned long > > keyword_index;
      compile_index( input_data,
                     keyword_index,
                     keywords );

      for( map< string, set< unsigned long > >::const_iterator it= keyword_index.begin();
           it!= keyword_index.end(); ++it )
        {
          cout << it->first << ", ";
          for( set< unsigned long >::const_iterator jt= it->second.begin(); 
               jt!= it->second.end(); ++jt )
            {
              if( jt!= it->second.begin() ) {
                cout << ", ";
              }
              if( page_labels_b ) {
                if( *jt< page_labels.size() ) {
                  cout << page_labels[ *jt ];
                }
                else { // error
                  cerr << "Error: page_labels index out of range" << endl;
                  cout << "ERROR";
                }
              }
              else {
                cout << (*jt+ 1); // translate our zero-base page ref. to one-based
              }
            }
          cout << endl;
        }

      cout << endl;
      success_b= true;
    }
    else { // error
      cerr << "Error: unable to open file: " << argv[1] << endl;
    }
  }

  if( !success_b ) {
    cerr << "page_refs 1.0, Create Index-like Page References from Keywords" << endl;
    cerr << endl;
    cerr << "   page_refs requires two or three arguments: " << endl;
    cerr << "      <doc. text fn> <keywords fn> [<doc. data fn>]" << endl;
    cerr << "   where: " << endl;
    cerr << "      <doc. text fn> is the name of a text file with formfeed (0x0c)" << endl;
    cerr << "         page delimiters, as created by xpdf's pdftotext, for example." << endl;
    cerr << "      <keywords fn> is the name of a text file of keywords, as " << endl;
    cerr << "         created by kw_catcher.  Use a hyphen (-) to take keywords" << endl;
    cerr << "         from stdin." << endl;
    cerr << "      <doc. data fn> is the name of a text file of document data," << endl;
    cerr << "         as created by pdftk's dump_data operation.  This is used" << endl;
    cerr << "         to adjust document page numbers." << endl;
    cerr << endl;
    cerr << "   See Also: kw_catcher, pdftk" << endl;
    cerr << endl;

    return 1;
  }
  else {
    return 0;
  }
}
